/*  int.h
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#ifndef REDUZE_INT_H
#define REDUZE_INT_H

#include <vector>
#include <list>
#include <set>
#include <map>
#include <sstream>
#include "functions.h"
#include "sector.h"
#include "ginac/basic.h" // for class GiNaC::ex and typedef GiNaC::exmap
#include "yamlconfigurable.h"

#ifdef USE_HASH_TABLE
#include <hash_map>
#include <hash_set>
#endif

namespace Reduze {

class IntegralFamily;
class SectorMappings;
class INT2;
class LinearCombination;

/// Feynman integral
/** An integral is a list of exponents, where positive (negative) entries
 *  belong to propagators in the denominator (numerator). ID is the
 *  identification number of the sector, t the number of the propagators
 *  in the denominator and r (s) the absolute value of the sum of the exponents
 *  in the denominator (numerator).
 */
class INT {
public:
	/// generates an integral with all values set to 0
	/** default INT is the trivial zero integral for #loop momenta > 0,
	 ** or defined as 1 for #loop momenta == 0 **/
	INT(const IntegralFamily* m);
	/// generates the integral of topology ID_ with no additional propagators
	explicit INT(const IntegralFamily* m, int id);
	/// generates the integral of topology ID_ with no additional propagators
	explicit INT(const Sector& s);
	/// generates an integral from the vector v_
	explicit INT(const IntegralFamily* m, const std::vector<int8> & v);
	/// generates an integral from a string in Reduze format
	explicit INT(const std::string & str);
	/// generates an integral from a YAML node
	explicit INT(const YAML::Node& node);
	/// generates an integral from an INT2 using the map to replace symbols by numbers, aborts if replacing fails
	explicit INT(const INT2& in2, const GiNaC::exmap& m);

	/// returns an INT constructed from Mathematica-type string
	static INT from_mma(const std::string& s);

	// getter/setter

	/// returns the IntegralFamily the INT belongs to
	const IntegralFamily* integralfamily() const;
	/// returns the sector the INT belongs to
	Sector get_sector() const;

	/// returns the length of the vector v (integralfamily()->get_NoP())
	size_t size() const {
		return v_.size();
	}
	/// sets the metric convention the integral is defined in
	/** b==false: g=diag(1,-1,-1,-1), b==true: g=diag(-1,1,1,1) **/
	void set_metric_mostly_plus(bool b) {
		metric_mostly_plus_ = b;
	}
	bool uses_metric_mostly_plus() const {
		return metric_mostly_plus_;
	}

	//	bool same_vector(const std::vector<int8> vec) const {
	//		return v_ == vec;
	//	}
	//	bool same_family(const IntegralFamily* ic) const {
	//		return integralfamily_ == ic;
	//	}

	/// returns the whole vector v
	std::vector<int8> v() const {
		return v_;
	}
	/// returns v[index]
	int8 v_i(int index) const {
		return v_[index];
	}
	/// returns the ID number
	int id() const {
		return id_;
	}
	/// returns t
	int8 t() const {
		return t_;
	}
	/// returns r
	int8 r() const {
		return r_;
	}
	/// returns s
	int8 s() const {
		return s_;
	}
	/// returns the maximal v_i
	int8 max_v_i() const;
	/// returns the minimal v_i
	int8 min_v_i() const;

	/// returns the minimal equivalent integral of integral 'integral'
	/** employs the permutation symmetries*/
	INT get_equivalent() const;

	/// applies (the inverse of) the permutation 'cycles' to the integral
	INT& transform(const std::vector<std::vector<int> >& cycles);

	/// applies the shift on the momenta of the propagators and tries to match the result with the integral family 'ic'
	// multiplies result with the integer 'mult'
	LinearCombination
			apply_shift(const GiNaC::exmap& shift, const IntegralFamily* ic,
					int mult, bool replace_symbol_to_replace_by_one,
					bool set_subsec_to_zero) const;

	// properties of the integral

	/// returns true if the integral is obviously zero, false otherwise
	bool is_obviously_zero() const;
	/// returns true if the integral is cut to zero, false otherwise
	/** returns true if a propagator, which is declared as a cut propagator
	 *	appears in the integral with an exponent <= 0 */
	bool is_cut_to_zero() const;
	/// returns the mass dimension of the integrand: 2*(s-r)
	int mass_dimension() const {
		return 2 * (static_cast<int> (s_) - static_cast<int> (r_));
	}
	/// returns the integrand in terms of instances of Propagator
	GiNaC::ex get_integrand() const;
	/// returns a hash value
	size_t hash_value() const;

	/// returns true if *this is equal to other (each member is equal) and false otherwise
	bool operator ==(const INT& other) const;
	/// returns true if operator==  returns false, false otherwise
	bool operator !=(const INT& other) const;
	/// returns true if *this is less complicated than other
	bool operator <(const INT& other) const {
		bool preferred = is_preferred(*this);
		if (preferred != is_preferred(other))
			return preferred;
		return less_rs(other); // return less_rs(other);
	}
	/// returns true if *this is more complicated than other
	bool operator >(const INT& other) const;

	// input/output

	///  writes the integral to the stream 'strm' in FORM format
	void to_form_stream(std::ostream & strm) const;
	///  writes the integral to the stream 'strm' in Maple format
	void to_maple_stream(std::ostream & strm) const;
	///  writes the integral to the stream 'strm' in Mathematica format
	void to_mma_stream(std::ostream & strm) const;
	/// format is a space separated row of the exponents
	friend std::ostream & operator <<(std::ostream& strm, const INT & in);
	/// reads from YAML
	friend void operator >>(const YAML::Node& node, INT& in);
	/// writes to YAML
	friend YAML::Emitter& operator <<(YAML::Emitter& os, const INT& in);

	/// sets integral and its crossed partners to be preferred masters
	/** change ordering: set integral lower than all which are not preferred */
	static void set_preferred(const INT&);
	/// set_preferred() for all integrals contained in a file
	/** empty filename is ignored **/
	static void load_preferred(const std::string& filename);
	/// whether integral or crossed partner is preferred master
	/** the integral with r = s = 0 is always preferred **/
	static bool is_preferred(const INT&);

private:
	const IntegralFamily* integralfamily_;
	/// the integer entries of the vector v are the exponents of the propagators
	std::vector<int8> v_;
	/// the ID number is an identifier for a topology
	int id_;
	/// t is the number of different propagators in the denominator
	int8 t_;
	/// r is the sum of the exponents in the denominator
	int8 r_;
	/// s is the sum of the exponents in the numerator, (absolut value)
	int8 s_;
	/// metric convention the integral is defined in, default: false
	/** b==false: g=diag(1,-1,-1,-1), b==true: g=diag(-1,1,1,1) **/
	bool metric_mostly_plus_;

	/// calculates the parameters t_, r_, s_, id_ from v_ and sets preferred = true for corner integral, false otherwise
	void init();

	// different orderings, ignoring 'preferred' feature
	bool less_rs(const INT& other) const;
	bool less_degrlex_r(const INT& other) const;
	bool less_degrlex_s(const INT& other) const;

	/// comparison class for set of preferred INTs
	struct CompareNoPreferred {
		bool operator()(const INT& a, const INT& b) {
			return a.less_rs(b);
		}
	};

	/// integrals which are preferred
	/** store only uncrossed INTs here */
	static std::set<INT, CompareNoPreferred> preferred_integrals;
};

// inline public members

inline const IntegralFamily* INT::integralfamily() const {
	return integralfamily_;
}
inline Sector INT::get_sector() const {
	return Sector(integralfamily_, id_);
}

inline bool INT::operator>(const INT& other) const {
	return other < *this;
}

// GLOBAL FUNCTIONS


/// operator for printing an integral
std::ostream & operator<<(std::ostream & strm, const INT & in);

template<class T> T string_to_T(const std::string & str);

class RSPoint;
class RSFiniteGenericSelection;

class SeedGeneratorOptions: public YAMLConfigurable {
public:
	static YAMLSpec yaml_spec() {
		YAMLSpec s;
		s.set_keyword("seed_generator_options");
		s.set_short_description("Options for the seed generator.");
		s.set_long_description(""//
					"Options for the seed generator.");
		s.add_option("reduce_to_equivalent", false, "boolean", ""//
					"Whether to replace integrals by lowest equivalent"
					" according to permutation symmetries.");
		return s;
	}
	virtual YAMLSpec yaml_spec_link() const {
		return yaml_spec();
	}

	SeedGeneratorOptions() :
		reduce_to_equivalent_(true) {
	}
	virtual ~SeedGeneratorOptions() {
	}

	/// whether permutation symmetries of the integralfamily should be used
	void set_reduce_to_equivalent(bool b);
	virtual void print(YAML::Emitter& os) const;
	virtual void read(const YAML::Node&);

private:
	bool reduce_to_equivalent_;
	friend class SeedGenerator;
};

// operators
inline void operator >>(const YAML::Node& node, SeedGeneratorOptions& opt) {
	opt.read(node);
}
inline YAML::Emitter& operator <<(YAML::Emitter& os,
		const SeedGeneratorOptions& opt) {
	opt.print(os);
	return os;
}

/// class to generate seed integrals
class SeedGenerator {
public:
	SeedGenerator(const IntegralFamily* ic, const SeedGeneratorOptions& options) :
		integralfamily_(ic), /**/
		options_(options) {
	}
	virtual ~SeedGenerator() {
	}

	/// generate all integrals of topology id and class (r,s)
	/** integrals are appended to out */
	void generate(int id, int r, int s, std::set<INT>& out) const;
	/// generate all integrals of topology id and with r in (r_min,r_max) and s in (s_min,s_max)
	/** integrals are appended to out */
	void generate(int id, int r_min, int r_max, int s_min, int s_max, std::set<
			INT>& out) const;
	/// generate all integrals of topology id which are contained in the range rs_generic
	/** integrals are appended to out */
	void generate(int id, const RSFiniteGenericSelection& rs_generic, std::set<
			INT>& out) const;
	/// generate all integrals of topology id which are contained in the ranges rs_generic
	/** integrals are appended to out */
	void generate(int id,
			const std::list<RSFiniteGenericSelection>& rs_generic,
			std::set<INT>& out) const;

private:
	/// never unbound
	SeedGenerator() :
		integralfamily_(0) {
	}

	const IntegralFamily* integralfamily_;
	SeedGeneratorOptions options_;
};

} // namespace Reduze


#ifdef USE_HASH_TABLE
namespace __gnu_cxx {
	template<> struct hash<Reduze::INT> {
		size_t operator()(const Reduze::INT& i) const {
			return i.hash_value();
		}
	};
} // namespace __gnu_cxx
#endif

#endif

